(ns net.serice.clojure.rational-to-string-test
  (use clojure.test)
  (use (clojure [pprint :only (cl-format)]))
  (use net.serice.clojure.rational-to-string)
  (:gen-class))

(defn harmonic-series [n]
  "Calculates the sum of the first N terms of the harmonic series:
    1 + 1/2 + 1/3 + ... + 1/n"
  (letfn
      ((iter [n acc]
         (if (<= n 0)
           acc
           (iter (- n 1) (+ (/ n) acc)))))
    (iter n 0)))

;;;
;;; (rational-to-string)
;;;

(deftest invalid-characters
  (is (= [nil "foo" 0]        (string-to-rational "foo")))
  (is (= [123 "x" 3]          (string-to-rational "123x")))
  (is (= [nil "x123" 0]       (string-to-rational "x123")))
  (is (= [1 "x3" 1]           (string-to-rational "1x3")))
  (is (= [nil "x.123" 0]      (string-to-rational "x.123")))
  (is (= [0 "x.123" 1]        (string-to-rational "0x.123")))
  (is (= [0 "x123" 2]         (string-to-rational "0.x123")))
  (is (= [(/ 123 1000) "x" 5] (string-to-rational "0.123x"))))

(deftest normalize-test

  ;;
  ;; x = 0
  ;;
  (is (= [0 0]             (normalize-rational  0)))

  ;;
  ;; x = 1
  ;;
  (is (= [1 0]             (normalize-rational  1)))
  ;;
  ;; abs(x) < 1
  ;;
  (is (= [7 -3]             (normalize-rational (/  7 1000))))
  (is (= [8 -3]             (normalize-rational (/  8 1000))))
  (is (= [9 -3]             (normalize-rational (/  9 1000))))
  (is (= [1 -2]             (normalize-rational (/ 10 1000))))
  (is (= [(/ 11 10) -2]     (normalize-rational (/ 11 1000))))
  (is (= [(/ 12 10) -2]     (normalize-rational (/ 12 1000))))
  (is (= [(/ 13 10) -2]     (normalize-rational (/ 13 1000))))

  (is (= [7 -2]             (normalize-rational (/  7 100))))
  (is (= [8 -2]             (normalize-rational (/  8 100))))
  (is (= [9 -2]             (normalize-rational (/  9 100))))
  (is (= [1 -1]             (normalize-rational (/ 10 100))))
  (is (= [(/ 11 10) -1]     (normalize-rational (/ 11 100))))
  (is (= [(/ 12 10) -1]     (normalize-rational (/ 12 100))))
  (is (= [(/ 13 10) -1]     (normalize-rational (/ 13 100))))

  (is (= [7 -1]             (normalize-rational (/  7 10))))
  (is (= [8 -1]             (normalize-rational (/  8 10))))
  (is (= [9 -1]             (normalize-rational (/  9 10))))
  (is (= [1 0]             (normalize-rational (/ 10 10))))
  (is (= [(/ 11 10) 0]      (normalize-rational (/ 11 10))))
  (is (= [(/ 12 10) 0]      (normalize-rational (/ 12 10))))
  (is (= [(/ 13 10) 0]      (normalize-rational (/ 13 10))))

  (is (= [-7 -3]            (normalize-rational (/  -7 1000))))
  (is (= [-8 -3]            (normalize-rational (/  -8 1000))))
  (is (= [-9 -3]            (normalize-rational (/  -9 1000))))
  (is (= [-1 -2]            (normalize-rational (/ -10 1000))))
  (is (= [(/ -11 10) -2]    (normalize-rational (/ -11 1000))))
  (is (= [(/ -12 10) -2]    (normalize-rational (/ -12 1000))))
  (is (= [(/ -13 10) -2]    (normalize-rational (/ -13 1000))))

  (is (= [-7 -2]            (normalize-rational (/  -7 100))))
  (is (= [-8 -2]            (normalize-rational (/  -8 100))))
  (is (= [-9 -2]            (normalize-rational (/  -9 100))))
  (is (= [-1 -1]            (normalize-rational (/ -10 100))))
  (is (= [(/ -11 10) -1]    (normalize-rational (/ -11 100))))
  (is (= [(/ -12 10) -1]    (normalize-rational (/ -12 100))))
  (is (= [(/ -13 10) -1]    (normalize-rational (/ -13 100))))

  (is (= [-7 -1]            (normalize-rational (/  -7 10))))
  (is (= [-8 -1]            (normalize-rational (/  -8 10))))
  (is (= [-9 -1]            (normalize-rational (/  -9 10))))
  (is (= [-1 0]            (normalize-rational (/ -10 10))))
  (is (= [(/ -11 10) 0]     (normalize-rational (/ -11 10))))
  (is (= [(/ -12 10) 0]     (normalize-rational (/ -12 10))))
  (is (= [(/ -13 10) 0]     (normalize-rational (/ -13 10))))

  ;;
  ;; abs(x) > 1
  ;;
  (is (= [7 0]              (normalize-rational  7)))
  (is (= [8 0]              (normalize-rational  8)))
  (is (= [9 0]              (normalize-rational  9)))
  (is (= [1 1]              (normalize-rational 10)))
  (is (= [(/ 11 10) 1]      (normalize-rational 11)))
  (is (= [(/ 12 10) 1]      (normalize-rational 12)))
  (is (= [(/ 13 10) 1]      (normalize-rational 13)))

  (is (= [(/ 97 10) 1]      (normalize-rational  97)))
  (is (= [(/ 98 10) 1]      (normalize-rational  98)))
  (is (= [(/ 99 10) 1]      (normalize-rational  99)))
  (is (= [1 2]              (normalize-rational 100)))
  (is (= [(/ 101 100) 2]    (normalize-rational 101)))
  (is (= [(/ 102 100) 2]    (normalize-rational 102)))
  (is (= [(/ 103 100) 2]    (normalize-rational 103)))

  (is (= [(/ 997 100) 2]    (normalize-rational  997)))
  (is (= [(/ 998 100) 2]    (normalize-rational  998)))
  (is (= [(/ 999 100) 2]    (normalize-rational  999)))
  (is (= [1 3]              (normalize-rational 1000)))
  (is (= [(/ 1001 1000) 3]  (normalize-rational 1001)))
  (is (= [(/ 1002 1000) 3]  (normalize-rational 1002)))
  (is (= [(/ 1003 1000) 3]  (normalize-rational 1003)))

  (is (= [-7 0]             (normalize-rational  -7)))
  (is (= [-8 0]             (normalize-rational  -8)))
  (is (= [-9 0]             (normalize-rational  -9)))
  (is (= [-1 1]             (normalize-rational -10)))
  (is (= [(/ -11 10) 1]     (normalize-rational -11)))
  (is (= [(/ -12 10) 1]     (normalize-rational -12)))
  (is (= [(/ -13 10) 1]     (normalize-rational -13)))

  (is (= [(/ -97 10) 1]     (normalize-rational  -97)))
  (is (= [(/ -98 10) 1]     (normalize-rational  -98)))
  (is (= [(/ -99 10) 1]     (normalize-rational  -99)))
  (is (= [-1 2]             (normalize-rational -100)))
  (is (= [(/ -101 100) 2]   (normalize-rational -101)))
  (is (= [(/ -102 100) 2]   (normalize-rational -102)))
  (is (= [(/ -103 100) 2]   (normalize-rational -103)))

  (is (= [(/ -997 100) 2]   (normalize-rational  -997)))
  (is (= [(/ -998 100) 2]   (normalize-rational  -998)))
  (is (= [(/ -999 100) 2]   (normalize-rational  -999)))
  (is (= [-1 3]             (normalize-rational -1000)))
  (is (= [(/ -1001 1000) 3] (normalize-rational -1001)))
  (is (= [(/ -1002 1000) 3] (normalize-rational -1002)))
  (is (= [(/ -1003 1000) 3] (normalize-rational -1003))))

(deftest negative-test
  (is (thrown? RuntimeException (rational-to-string -1 -1 :trim true)))

  (is (= "-1"      (rational-to-string -1 0 :trim true)))
  (is (= "-1"      (rational-to-string -1 1 :trim true)))
  (is (= "-1"      (rational-to-string -1 2 :trim true)))
  (is (= "-1"      (rational-to-string -1 3 :trim true)))
  (is (= "-1"      (rational-to-string -1 4 :trim true)))

  (is (= "-1"      (rational-to-string -1 0 :trim nil)))
  (is (= "-1.0"    (rational-to-string -1 1 :trim nil)))
  (is (= "-1.00"   (rational-to-string -1 2 :trim nil)))
  (is (= "-1.000"  (rational-to-string -1 3 :trim nil)))
  (is (= "-1.0000" (rational-to-string -1 4 :trim nil)))

  (is (= "-1"      (rational-to-string (/ -10001 10000) 0 :trim true)))
  (is (= "-1"      (rational-to-string (/ -10001 10000) 1 :trim true)))
  (is (= "-1"      (rational-to-string (/ -10001 10000) 2 :trim true)))
  (is (= "-1"      (rational-to-string (/ -10001 10000) 3 :trim true)))
  (is (= "-1.0001" (rational-to-string (/ -10001 10000) 4 :trim true)))

  (is (= "-1"      (rational-to-string (/ -10001 10000) 0 :trim nil)))
  (is (= "-1.0"    (rational-to-string (/ -10001 10000) 1 :trim nil)))
  (is (= "-1.00"   (rational-to-string (/ -10001 10000) 2 :trim nil)))
  (is (= "-1.000"  (rational-to-string (/ -10001 10000) 3 :trim nil)))
  (is (= "-1.0001" (rational-to-string (/ -10001 10000) 4 :trim nil))))

(deftest zero-test
  (is (thrown? RuntimeException (rational-to-string 0 -1 :trim true)))
  (is (= "0"     (rational-to-string 0  0 :trim true)))
  (is (= "0"     (rational-to-string 0  1 :trim true)))
  (is (= "0"     (rational-to-string 0  2 :trim true)))
  (is (= "0"     (rational-to-string 0  3 :trim true)))
  (is (= "0"     (rational-to-string 0  0 :trim nil)))
  (is (= "0.0"   (rational-to-string 0  1 :trim nil)))
  (is (= "0.00"  (rational-to-string 0  2 :trim nil)))
  (is (= "0.000" (rational-to-string 0  3 :trim nil))))

(deftest positive-test
  (is (thrown? RuntimeException (rational-to-string -1 -1 :trim true)))

  (is (= "1"      (rational-to-string 1 0 :trim true)))
  (is (= "1"      (rational-to-string 1 1 :trim true)))
  (is (= "1"      (rational-to-string 1 2 :trim true)))
  (is (= "1"      (rational-to-string 1 3 :trim true)))
  (is (= "1"      (rational-to-string 1 4 :trim true)))

  (is (= "1"      (rational-to-string 1 0 :trim nil)))
  (is (= "1.0"    (rational-to-string 1 1 :trim nil)))
  (is (= "1.00"   (rational-to-string 1 2 :trim nil)))
  (is (= "1.000"  (rational-to-string 1 3 :trim nil)))
  (is (= "1.0000" (rational-to-string 1 4 :trim nil)))

  (is (= "1"      (rational-to-string (/ 10001 10000) 0 :trim nil)))
  (is (= "1.0"    (rational-to-string (/ 10001 10000) 1 :trim nil)))
  (is (= "1.00"   (rational-to-string (/ 10001 10000) 2 :trim nil)))
  (is (= "1.000"  (rational-to-string (/ 10001 10000) 3 :trim nil)))
  (is (= "1.0001" (rational-to-string (/ 10001 10000) 4 :trim nil))))

(deftest no-trim-test
  (is (= "1"       (rational-to-string (/ 1001 1000) 0 :trim nil)))
  (is (= "1.0"     (rational-to-string (/ 1001 1000) 1 :trim nil)))
  (is (= "1.00"    (rational-to-string (/ 1001 1000) 2 :trim nil)))
  (is (= "1.001"   (rational-to-string (/ 1001 1000) 3 :trim nil)))
  (is (= "1.0010"  (rational-to-string (/ 1001 1000) 4 :trim nil)))
  (is (= "1.00100" (rational-to-string (/ 1001 1000) 5 :trim nil))))

(deftest trim-test
  (is (= "1"     (rational-to-string (/ 1001 1000) 0 :trim true)))
  (is (= "1"     (rational-to-string (/ 1001 1000) 1 :trim true)))
  (is (= "1"     (rational-to-string (/ 1001 1000) 2 :trim true)))
  (is (= "1.001" (rational-to-string (/ 1001 1000) 3 :trim true)))
  (is (= "1.001" (rational-to-string (/ 1001 1000) 4 :trim true)))
  (is (= "1.001" (rational-to-string (/ 1001 1000) 5 :trim true))))

(deftest rounding-test
  (is (= "0"       (rational-to-string (/ 1 100000) 0 :trim true)))
  (is (= "0"       (rational-to-string (/ 1 100000) 1 :trim true)))
  (is (= "0"       (rational-to-string (/ 1 100000) 2 :trim true)))
  (is (= "0"       (rational-to-string (/ 1 100000) 3 :trim true)))
  (is (= "0"       (rational-to-string (/ 1 100000) 4 :trim true)))
  (is (= "0.00001" (rational-to-string (/ 1 100000) 5 :trim true)))
  (is (= "0.00001" (rational-to-string (/ 1 100000) 6 :trim true)))
  (is (= "0.00001" (rational-to-string (/ 1 100000) 6 :trim true)))

  (is (= "0"       (rational-to-string (/ 1 10) 0 :trim true)))
  (is (= "0.1"     (rational-to-string (/ 1 10) 1 :trim true)))
  (is (= "0.1"     (rational-to-string (/ 1 10) 2 :trim true)))

  (is (= "0"       (rational-to-string (/ 49999 100000) 0 :trim true)))
  (is (= "0.5"     (rational-to-string (/ 49999 100000) 1 :trim true)))
  (is (= "0.5"     (rational-to-string (/ 49999 100000) 2 :trim true)))
  (is (= "0.5"     (rational-to-string (/ 49999 100000) 3 :trim true)))
  (is (= "0.5"     (rational-to-string (/ 49999 100000) 4 :trim true)))
  (is (= "0.49999" (rational-to-string (/ 49999 100000) 5 :trim true)))

  (is (= "2"       (rational-to-string (/ 3 2) 0 :trim true)))
  (is (= "1.5"     (rational-to-string (/ 3 2) 1 :trim true)))
  (is (= "1.5"     (rational-to-string (/ 3 2) 2 :trim true)))

  (is (= "100000"  (rational-to-string (/ 199999 2) 0 :trim true)))
  (is (= "99999.5" (rational-to-string (/ 199999 2) 1 :trim true)))
  (is (= "99999.5" (rational-to-string (/ 199999 2) 2 :trim true)))

  (is (= "1.235"   (rational-to-string (/ 12345 10000) 3 :trim true)))
  (is (= "-1.235"  (rational-to-string (/ -12345 10000) 3 :trim true)))
  (is (= "1.2346"  (rational-to-string (/ 123455 100000) 4 :trim true)))
  (is (= "-1.2346" (rational-to-string (/ -123455 100000) 4 :trim true))))

(deftest decimal-placement
  (let [expt
        @#'net.serice.clojure.rational-to-string/expt]
    (is (= "0.00001" (rational-to-string (expt 10 -5) 10 :trim true)))
    (is (=  "0.0001" (rational-to-string (expt 10 -4) 10 :trim true)))
    (is (=   "0.001" (rational-to-string (expt 10 -3) 10 :trim true)))
    (is (=    "0.01" (rational-to-string (expt 10 -2) 10 :trim true)))
    (is (=     "0.1" (rational-to-string (expt 10 -1) 10 :trim true)))
    (is (=       "1" (rational-to-string (expt 10  0) 10 :trim true)))
    (is (=      "10" (rational-to-string (expt 10  1) 10 :trim true)))
    (is (=     "100" (rational-to-string (expt 10  2) 10 :trim true)))
    (is (=    "1000" (rational-to-string (expt 10  3) 10 :trim true)))
    (is (=   "10000" (rational-to-string (expt 10  4) 10 :trim true)))
    (is (=  "100000" (rational-to-string (expt 10  5) 10 :trim true)))))

(deftest harmonic-series-test
  "Test (harmonic-series 250)."
  (is (= "6.1006752494325792775723270159741508963779612275122648211523647985"
         (rational-to-string (harmonic-series 250) 64))))

(deftest rational-to-string-test
  (is (= "0.02" (rational-to-string (/ 15 1000) 2 :sci nil :trim nil)))
  (is (= "0.002" (rational-to-string (/ 15 10000) 3 :sci nil :trim nil)))
)

;;;
;;; (string-to-rational)
;;;

(deftest is-white-space-test
  (let [is-white-space
        @#'net.serice.clojure.rational-to-string/is-white-space]
    (is (= true (is-white-space \space)))
    (is (= true (is-white-space \tab)))
    (is (= true (is-white-space \newline)))
    (is (= true (is-white-space \return)))
    (is (= nil (is-white-space \a)))))

(deftest skip-white-space-test
  (let [skip-white-space
        @#'net.serice.clojure.rational-to-string/skip-white-space]
    (is (= [(seq "")    0] (skip-white-space (seq ""))))
    (is (= [(seq "")    1] (skip-white-space (seq " "))))
    (is (= [(seq "")    2] (skip-white-space (seq "  "))))
    (is (= [(seq "")    3] (skip-white-space (seq "   "))))
    (is (= [(seq "1")   0] (skip-white-space (seq "1"))))
    (is (= [(seq "12")  0] (skip-white-space (seq "12"))))
    (is (= [(seq "123") 0] (skip-white-space (seq "123"))))
    (is (= [(seq "1")   0] (skip-white-space (seq "1"))))
    (is (= [(seq "12")  0] (skip-white-space (seq "12"))))
    (is (= [(seq "123") 0] (skip-white-space (seq "123"))))
    (is (= [(seq "1")   1] (skip-white-space (seq " 1"))))
    (is (= [(seq "12")  1] (skip-white-space (seq " 12"))))
    (is (= [(seq "123") 1] (skip-white-space (seq " 123"))))
    (is (= [(seq "1")   2] (skip-white-space (seq "  1"))))
    (is (= [(seq "12")  2] (skip-white-space (seq "  12"))))
    (is (= [(seq "123") 2] (skip-white-space (seq "  123"))))))

(deftest is-decimal-digit-test
  (let [is-decimal-digit
        @#'net.serice.clojure.rational-to-string/is-decimal-digit]
    (is (= true (is-decimal-digit \0)))
    (is (= true (is-decimal-digit \1)))
    (is (= true (is-decimal-digit \2)))
    (is (= true (is-decimal-digit \3)))
    (is (= true (is-decimal-digit \4)))
    (is (= true (is-decimal-digit \5)))
    (is (= true (is-decimal-digit \6)))
    (is (= true (is-decimal-digit \7)))
    (is (= true (is-decimal-digit \8)))
    (is (= true (is-decimal-digit \9)))
    (is (= nil (is-decimal-digit \a)))))

(deftest expt-test
  (let [expt
        @#'net.serice.clojure.rational-to-string/expt]
    (is (= (/ 1 1000) (expt 10 -3)))
    (is (= (/ 1 100) (expt 10 -2)))
    (is (= (/ 1 10) (expt 10 -1)))
    (is (= 1 (expt 10 0)))
    (is (= 10 (expt 10 1)))
    (is (= 100 (expt 10 2)))
    (is (= 1000 (expt 10 3)))))

(deftest cs-to-sign-test
  (is (= [nil nil 0] (cs-to-sign (seq ""))))
  (is (= [nil (seq "1") 0] (cs-to-sign (seq "1"))))
  (is (= [:plus (seq "1") 1] (cs-to-sign (seq "+1"))))
  (is (= [:minus (seq "1") 1] (cs-to-sign (seq "-1")))))

(deftest cs-to-uint-test
  (is (= [nil nil 0] (cs-to-uint (seq ""))))
  (is (= [1N nil 1] (cs-to-uint (seq "1"))))
  (is (= [1N nil 2] (cs-to-uint (seq "01"))))
  (is (= [nil (seq "+") 0] (cs-to-uint (seq "+"))))
  (is (= [nil (seq "+1") 0] (cs-to-uint (seq "+1"))))
  (is (= [nil (seq "-") 0] (cs-to-uint (seq "-"))))
  (is (= [nil (seq "-1") 0] (cs-to-uint (seq "-1")))))

(deftest cs-to-int-test
  (is (= [nil nil 0] (cs-to-int (seq ""))))
  (is (= [1N nil 1] (cs-to-int (seq "1"))))
  (is (= [1N nil 2] (cs-to-int (seq "01"))))
  (is (= [nil (seq "+") 0] (cs-to-int (seq "+"))))
  (is (= [1N nil 2] (cs-to-int (seq "+1"))))
  (is (= [1N nil 3] (cs-to-int (seq "+01"))))
  (is (= [nil (seq "-") 0] (cs-to-int (seq "-"))))
  (is (= [-1N nil 3] (cs-to-int (seq "-01")))))

(deftest cs-to-fract-test
  (is (= [nil nil 0] (cs-to-fract (seq ""))))
  (is (= [nil (seq "1") 0] (cs-to-fract (seq "1"))))
  (is (= [nil nil 1] (cs-to-fract (seq "."))))
  (is (= [1N nil 2] (cs-to-fract (seq ".1"))))
  (is (= [1N nil 3] (cs-to-fract (seq ".01"))))
  (is (= [1N nil 4] (cs-to-fract (seq ".001"))))
  (is (= [nil (seq "+1") 1] (cs-to-fract (seq ".+1"))))
  (is (= [nil (seq "-1") 1] (cs-to-fract (seq ".-1"))))
  (is (= [nil (seq "a") 1] (cs-to-fract (seq ".a")))))

(deftest cs-to-expt-test
  (is (= [nil nil 0] (cs-to-expt (seq ""))))
  (is (= [nil (seq "1") 0] (cs-to-expt (seq "1"))))
  (is (= [nil (seq "e") 0] (cs-to-expt (seq "e"))))
  (is (= [1N nil 2] (cs-to-expt (seq "e1"))))
  (is (= [1N nil 3] (cs-to-expt (seq "e01"))))
  (is (= [1N nil 4] (cs-to-expt (seq "e001")))))

(deftest cs-to-decomposed-float-test
  (is (= [[nil nil nil     nil] nil 0]
         (cs-to-decomposed-float (seq ""))))

  (is (= [[nil   1N   [nil 0] nil] nil 1]
         (cs-to-decomposed-float (seq "1"))))

  (is (= [[nil   1N   [nil 0] nil] nil 2]
         (cs-to-decomposed-float (seq "1."))))

  (is (= [[nil   1N   [2N  1] nil] nil 3]
         (cs-to-decomposed-float (seq "1.2"))))

  (is (= [[nil   1N   [2N  1] nil] (seq "e") 3]
         (cs-to-decomposed-float (seq "1.2e"))))

  (is (= [[nil   1N   [2N  1] 3N ] nil 5]
         (cs-to-decomposed-float (seq "1.2e3"))))

  (is (= [[nil   nil  [2N  1] 3N ] nil 4]
         (cs-to-decomposed-float (seq ".2e3"))))

  (is (= [[nil   1N   [nil 0] 3N ] nil 3]
         (cs-to-decomposed-float (seq "1e3"))))

  (is (= [[nil   1N   [nil 0] 3N ] nil 4]
         (cs-to-decomposed-float (seq "1.e3"))))

  (is (= [[:plus 1N   [nil 0] nil] nil 2]
         (cs-to-decomposed-float (seq "+1"))))

  (is (= [[:minus 1N  [nil 0] nil] nil 2]
         (cs-to-decomposed-float (seq "-1"))))

  (is (= [[:plus 1N   [nil 0] nil] nil 3]
         (cs-to-decomposed-float (seq "+1."))))

  (is (= [[:minus 1N  [nil 0] nil] nil 3]
         (cs-to-decomposed-float (seq "-1."))))

  (is (= [[:plus 1N   [2N  1] nil] nil 4]
         (cs-to-decomposed-float (seq "+1.2"))))

  (is (= [[:minus 1N  [2N  1] nil] nil 4]
         (cs-to-decomposed-float (seq "-1.2"))))

  (is (= [[:plus 1N   [2N  1] 3N ] nil 6]
         (cs-to-decomposed-float (seq "+1.2e3"))))

  (is (= [[:minus 1N  [2N  1] 3N ] nil 6]
         (cs-to-decomposed-float (seq "-1.2e3"))))

  (is (= [[:plus 1N   [2N  1] 3N ] nil 7]
         (cs-to-decomposed-float (seq "+1.2e+3"))))

  (is (= [[:minus 1N  [2N  1] -3N] nil 7]
         (cs-to-decomposed-float (seq "-1.2e-3"))))

  (is (= [[:plus 1N   [nil 0] 3N ] nil 4]
         (cs-to-decomposed-float (seq "+1e3"))))

  (is (= [[:minus 1N  [nil 0] 3N ] nil 4]
         (cs-to-decomposed-float (seq "-1e3"))))

  (is (= [[:plus 1N   [nil 0] 3N ] nil 5]
         (cs-to-decomposed-float (seq "+1.e3"))))

  (is (= [[:minus 1N  [nil 0] 3N ] nil 5]
         (cs-to-decomposed-float (seq "-1.e3"))))

  (is (= [[:plus 1N   [nil 0] 3N ] nil 5]
         (cs-to-decomposed-float (seq "+1e+3"))))

  (is (= [[:minus 1N  [nil 0] -3N] nil 5]
         (cs-to-decomposed-float (seq "-1e-3"))))

  (is (= [[:plus 1N   [nil 0] 3N ] nil 6]
         (cs-to-decomposed-float (seq "+1.e+3"))))

  (is (= [[:minus 1N  [nil 0] -3N] nil 6]
         (cs-to-decomposed-float (seq "-1.e-3")))))

(deftest cs-to-rational-test
  (is (= [nil nil 0]
         (cs-to-rational (seq ""))))

  (is (= [nil (seq ".") 0]
         (cs-to-rational (seq "."))))

  (is (= [nil (seq "e") 0]
         (cs-to-rational (seq "e"))))

  (is (= [nil (seq "e3") 0]
         (cs-to-rational (seq "e3"))))

  (is (= [1N nil 1]
         (cs-to-rational (seq "1"))))

  (is (= [1N nil 2]
         (cs-to-rational (seq "1."))))

  (is (= [(/ 2 10) nil 2]
         (cs-to-rational (seq ".2"))))

  (is (= [(+ 1 (/ 2 10)) nil 3]
         (cs-to-rational (seq "1.2"))))

  (is (= [(+ 1 (/ 2 10)) (seq "e") 3]
         (cs-to-rational (seq "1.2e"))))

  (is (= [(* (+ 1 (/ 2 10)) 1000) nil 5]
         (cs-to-rational (seq "1.2e3"))))

  (is (= [(* (+ 0 (/ 2 10)) 1000) nil 4]
         (cs-to-rational (seq ".2e3"))))

  (is (= [1N nil 2]
         (cs-to-rational (seq "+1"))))

  (is (= [-1N nil 2]
         (cs-to-rational (seq "-1"))))

  (is (= [1N nil 3]
         (cs-to-rational (seq "+1."))))

  (is (= [-1N nil 3]
         (cs-to-rational (seq "-1."))))

  (is (= [(/ 2 10) nil 3]
         (cs-to-rational (seq "+.2"))))

  (is (= [(- (/ 2 10)) nil 3]
         (cs-to-rational (seq "-.2"))))

  (is (= [(+ 1 (/ 2 10)) nil 4]
         (cs-to-rational (seq "+1.2"))))

  (is (= [(- (+ 1 (/ 2 10))) nil 4]
         (cs-to-rational (seq "-1.2"))))

  (is (= [(* (+ 1 (/ 2 10)) 1000) nil 6]
         (cs-to-rational (seq "+1.2e3"))))

  (is (= [(- (* (+ 1 (/ 2 10)) 1000)) nil 6]
         (cs-to-rational (seq "-1.2e3"))))

  (is (= [(* (+ 0 (/ 2 10)) 1000) nil 5]
         (cs-to-rational (seq "+.2e3"))))

  (is (= [(- (* (+ 0 (/ 2 10)) 1000)) nil 5]
         (cs-to-rational (seq "-.2e3"))))

  (is (= [(* (+ 1 (/ 2 10)) 1000) nil 7]
         (cs-to-rational (seq "+1.2e+3"))))

  (is (= [(/ (+ 1 (/ 2 10)) 1000) nil 7]
         (cs-to-rational (seq "+1.2e-3"))))

  (is (= [(- (* (+ 1 (/ 2 10)) 1000)) nil 7]
         (cs-to-rational (seq "-1.2e+3"))))

  (is (= [(- (/ (+ 1 (/ 2 10)) 1000)) nil 7]
         (cs-to-rational (seq "-1.2e-3"))))

  (is (= [(* (+ 0 (/ 2 10)) 1000) nil 6]
         (cs-to-rational (seq "+.2e+3"))))

  (is (= [(/ (+ 0 (/ 2 10)) 1000) nil 6]
         (cs-to-rational (seq "+.2e-3"))))

  (is (= [(- (* (+ 0 (/ 2 10)) 1000)) nil 6]
         (cs-to-rational (seq "-.2e+3"))))

  (is (= [(- (/ (+ 0 (/ 2 10)) 1000)) nil 6]
         (cs-to-rational (seq "-.2e-3")))))

(deftest invalid-characters
  (is (= [nil "foo" 0] (string-to-rational "foo")))
  (is (= [123 "x" 3] (string-to-rational "123x")))
  (is (= [nil "x123" 0] (string-to-rational "x123")))
  (is (= [1 "x3" 1] (string-to-rational "1x3")))
  (is (= [nil "x.123" 0] (string-to-rational "x.123")))
  (is (= [0 "x.123" 1] (string-to-rational "0x.123")))
  (is (= [0 "x123" 2] (string-to-rational "0.x123")))
  (is (= [(/ 123 1000) "x" 5] (string-to-rational "0.123x"))))

(deftest string-to-rational-test
  (is (= [nil "" 0] (string-to-rational "")))

  (is (= [nil "." 0] (string-to-rational ".")))
  (is (= [nil "e" 0] (string-to-rational "e")))

  (is (= [1N "" 1] (string-to-rational "1")))
  (is (= [1N "" 2] (string-to-rational "01")))
  (is (= [1N "" 3] (string-to-rational "001")))
  (is (= [1N "" 4] (string-to-rational "0001")))

  (is (= [1N "" 2] (string-to-rational "1.")))
  (is (= [1N "" 3] (string-to-rational "1.0")))
  (is (= [1N "" 4] (string-to-rational "1.00")))
  (is (= [1N "" 5] (string-to-rational "1.000")))

  (is (= [(+ 1 (/ 2 10)) "" 3] (string-to-rational "1.2")))
  (is (= [(+ 1 (/ 2 10)) "e" 3] (string-to-rational "1.2e")))
  (is (= [(* (+ 1 (/ 2 10)) 1000) "" 5] (string-to-rational "1.2e3")))
  (is (= [(* (+ 1 (/ 2 10)) 1000) "" 5] (string-to-rational "1.2E3")))

  (is (= [1N "" 2] (string-to-rational "+1")))
  (is (= [-1N "" 2] (string-to-rational "-1")))
  (is (= [1N "" 3] (string-to-rational "+1.")))
  (is (= [-1N "" 3] (string-to-rational "-1.")))

  (is (= [(/ 2 10) "" 3] (string-to-rational "+.2")))
  (is (= [(- (/ 2 10)) "" 3] (string-to-rational "-.2")))

  (is (= [(+ 1 (/ 2 10)) "" 4] (string-to-rational "+1.2")))
  (is (= [(- (+ 1 (/ 2 10))) "" 4] (string-to-rational "-1.2")))

  (is (= [1N "+2" 2] (string-to-rational "1.+2")))
  (is (= [1N "-2" 2] (string-to-rational "1.-2")))

  (is (= [(* (+ 1 (/ 2 10)) 1000) "" 6] (string-to-rational "+1.2e3")))
  (is (= [(- (* (+ 1 (/ 2 10)) 1000)) "" 6] (string-to-rational "-1.2e3")))
  (is (= [(* (+ 1 (/ 2 10)) 1000) "" 6] (string-to-rational "1.2e+3")))
  (is (= [(/ (+ 1 (/ 2 10)) 1000) "" 6] (string-to-rational "1.2e-3")))
  (is (= [(* (+ 1 (/ 2 10)) 1000) "" 7] (string-to-rational "+1.2e+3")))
  (is (= [(- (/ (+ 1 (/ 2 10)) 1000)) "" 7] (string-to-rational "-1.2e-3"))))

(deftest leading-implicit-positive-sign-test
  (is (= [nil "." 0] (string-to-rational ".")))
  (is (= [nil "..1" 0] (string-to-rational "..1")))
  (is (= [0 ".1" 2] (string-to-rational "0..1")))
  (is (= [(/ 10) "" 2] (string-to-rational ".1")))
  (is (= [1 "" 1] (string-to-rational "1")))
  (is (= [0 "" 1] (string-to-rational "0")))
  (is (= [0 "" 2] (string-to-rational "00")))
  (is (= [0 "" 3] (string-to-rational "000")))
  (is (= [0 "" 4] (string-to-rational "0000")))
  (is (= [(/ 10) "" 2] (string-to-rational ".1")))
  (is (= [(/ 10) "" 3] (string-to-rational "0.1")))
  (is (= [(/ 10) "" 4] (string-to-rational "00.1")))
  (is (= [(/ 12 100) "" 3] (string-to-rational ".12")))
  (is (= [(/ 12 100) "" 4] (string-to-rational "0.12")))
  (is (= [(/ 12 100) "" 5] (string-to-rational "00.12")))
  (is (= [1 "" 1] (string-to-rational "1")))
  (is (= [1 "" 2] (string-to-rational "01")))
  (is (= [1 "" 4] (string-to-rational " 001")))
  (is (= [1 "" 3] (string-to-rational "001")))
  (is (= [1 "" 4] (string-to-rational "0001")))
  (is (= [1 "" 5] (string-to-rational "00001"))))

(deftest leading-positive-sign-test
  (is (= [nil "+." 0] (string-to-rational "+.")))
  (is (= [nil ".+" 0] (string-to-rational ".+")))
  (is (= [1 "+" 1] (string-to-rational "1+")))
  (is (= [nil ".+1" 0] (string-to-rational ".+1")))
  (is (= [(/ 10) "+" 2] (string-to-rational ".1+")))
  (is (= [(/ 10) " +" 2] (string-to-rational ".1 +")))
  (is (= [nil "+..1" 0] (string-to-rational "+..1")))
  (is (= [0 ".1" 3] (string-to-rational "+0..1")))
  (is (= [nil "++1" 0] (string-to-rational "++1")))
  (is (= [0 "" 2] (string-to-rational "+0")))
  (is (= [0 "" 3] (string-to-rational "+00")))
  (is (= [0 "" 4] (string-to-rational "+000")))
  (is (= [0 "" 5] (string-to-rational "+0000")))
  (is (= [(/ 10) "" 3] (string-to-rational "+.1")))
  (is (= [(/ 10) "" 4] (string-to-rational "+0.1")))
  (is (= [(/ 10) "" 5] (string-to-rational "+00.1")))
  (is (= [(/ 12 100) "" 4] (string-to-rational "+.12")))
  (is (= [(/ 12 100) "" 5] (string-to-rational "+0.12")))
  (is (= [(/ 12 100) "" 6] (string-to-rational "+00.12")))
  (is (= [1 "" 2] (string-to-rational "+1")))
  (is (= [1 "" 3] (string-to-rational "+01")))
  (is (= [1 "" 5] (string-to-rational " +001")))
  (is (= [1 "" 4] (string-to-rational "+001")))
  (is (= [1 "" 5] (string-to-rational "+0001")))
  (is (= [1 "" 6] (string-to-rational "+00001"))))

(deftest leading-negative-sign-test
  (is (= [nil "-." 0] (string-to-rational "-.")))
  (is (= [nil ".-" 0] (string-to-rational ".-")))
  (is (= [1 "-" 1] (string-to-rational "1-")))
  (is (= [nil ".-1" 0] (string-to-rational ".-1")))
  (is (= [(/ 10) "-" 2] (string-to-rational ".1-")))
  (is (= [(/ 10) " -" 2] (string-to-rational ".1 -")))
  (is (= [nil ".-1" 0] (string-to-rational ".-1")))
  (is (= [nil "-..1" 0] (string-to-rational "-..1")))
  (is (= [0 ".1" 3] (string-to-rational "-0..1")))
  (is (= [nil "--1" 0] (string-to-rational "--1")))
  (is (= [0 "" 2] (string-to-rational "-0")))
  (is (= [0 "" 3] (string-to-rational "-00")))
  (is (= [0 "" 4] (string-to-rational "-000")))
  (is (= [0 "" 5] (string-to-rational "-0000")))
  (is (= [(/ -10) "" 3] (string-to-rational "-.1")))
  (is (= [(/ -10) "" 4] (string-to-rational "-0.1")))
  (is (= [(/ -10) "" 5] (string-to-rational "-00.1")))
  (is (= [(/ -12 100) "" 4] (string-to-rational "-.12")))
  (is (= [(/ -12 100) "" 5] (string-to-rational "-0.12")))
  (is (= [(/ -12 100) "" 6] (string-to-rational "-00.12")))
  (is (= [-1 "" 2] (string-to-rational "-1")))
  (is (= [-1 "" 3] (string-to-rational "-01")))
  (is (= [-1 "" 5] (string-to-rational " -001")))
  (is (= [-1 "" 4] (string-to-rational "-001")))
  (is (= [-1 "" 5] (string-to-rational "-0001")))
  (is (= [-1 "" 6] (string-to-rational "-00001"))))

(deftest space-test
  ;; empty strings
  (is (= [nil "" 0] (string-to-rational "")))
  (is (= [nil " " 0] (string-to-rational " ")))
  (is (= [nil "  " 0] (string-to-rational "  ")))
  (is (= [nil "   " 0] (string-to-rational "   ")))

  ;; leading space only
  (is (= [1 "" 4] (string-to-rational "   1")))
  (is (= [1 "" 5] (string-to-rational "   1.")))
  (is (= [12 "" 5] (string-to-rational "   12")))
  (is (= [12 "" 6] (string-to-rational "   12.")))
  (is (= [123 "" 6] (string-to-rational "   123")))
  (is (= [123 "" 7] (string-to-rational "   123.")))
  (is (= [(+ 123 (/ 4 10)) "" 8] (string-to-rational "   123.4")))
  (is (= [(+ 123 (/ 45 100)) "" 9] (string-to-rational "   123.45")))

  (is (= [1 "+" 4] (string-to-rational "   1+")))
  (is (= [1 "+" 5] (string-to-rational "   1.+")))
  (is (= [1 " +" 5] (string-to-rational "   1. +")))
  (is (= [1 "" 5] (string-to-rational "   +1")))
  (is (= [1 "" 6] (string-to-rational "   +1.")))
  (is (= [12 "" 6] (string-to-rational "   +12")))
  (is (= [12 "" 7] (string-to-rational "   +12.")))
  (is (= [123 "" 7] (string-to-rational "   +123")))
  (is (= [123 "" 8] (string-to-rational "   +123.")))
  (is (= [(+ 123 (/ 4 10)) "" 9] (string-to-rational "   +123.4")))
  (is (= [(+ 123 (/ 45 100)) "" 10] (string-to-rational "   +123.45")))

  (is (= [1 "-" 4] (string-to-rational "   1-")))
  (is (= [1 "-" 5] (string-to-rational "   1.-")))
  (is (= [1 " -" 5] (string-to-rational "   1. -")))
  (is (= [-1 "" 5] (string-to-rational "   -1")))
  (is (= [-1 "" 6] (string-to-rational "   -1.")))
  (is (= [-12 "" 6] (string-to-rational "   -12")))
  (is (= [-12 "" 7] (string-to-rational "   -12.")))
  (is (= [-123 "" 7] (string-to-rational "   -123")))
  (is (= [-123 "" 8] (string-to-rational "   -123.")))
  (is (= [(- -123 (/ 4 10)) "" 9] (string-to-rational "   -123.4")))
  (is (= [(- -123 (/ 45 100)) "" 10] (string-to-rational "   -123.45")))

  ; both leading and trailing space
  (is (= [1 "   " 4] (string-to-rational "   1   ")))
  (is (= [1 "   " 5] (string-to-rational "   1.   ")))
  (is (= [12 "   " 5] (string-to-rational "   12   ")))
  (is (= [12 "   " 6] (string-to-rational "   12.   ")))
  (is (= [123 "   " 6] (string-to-rational "   123   ")))
  (is (= [123 "   " 7] (string-to-rational "   123.   ")))
  (is (= [(+ 123 (/ 4 10)) "   " 8] (string-to-rational "   123.4   ")))
  (is (= [(+ 123 (/ 45 100)) "   " 9] (string-to-rational "   123.45   ")))

  (is (= [1 "+   " 4] (string-to-rational "   1+   ")))
  (is (= [1 "+   " 5] (string-to-rational "   1.+   ")))
  (is (= [1 " +   " 5] (string-to-rational "   1. +   ")))
  (is (= [1 "   " 5] (string-to-rational "   +1   ")))
  (is (= [1 "   " 6] (string-to-rational "   +1.   ")))
  (is (= [12 "   " 6] (string-to-rational "   +12   ")))
  (is (= [12 "   " 7] (string-to-rational "   +12.   ")))
  (is (= [123 "   " 7] (string-to-rational "   +123   ")))
  (is (= [123 "   " 8] (string-to-rational "   +123.   ")))
  (is (= [(+ 123 (/ 4 10)) "   " 9] (string-to-rational "   +123.4   ")))
  (is (= [(+ 123 (/ 45 100)) "   " 10] (string-to-rational "   +123.45   ")))

  (is (= [1 "-   " 4] (string-to-rational "   1-   ")))
  (is (= [1 "-   " 5] (string-to-rational "   1.-   ")))
  (is (= [1 " -   " 5] (string-to-rational "   1. -   ")))
  (is (= [-1 " " 5] (string-to-rational "   -1 ")))
  (is (= [-1 "   " 6] (string-to-rational "   -1.   ")))
  (is (= [-12 "   " 6] (string-to-rational "   -12   ")))
  (is (= [-12 "   " 7] (string-to-rational "   -12.   ")))
  (is (= [-123 "   " 7] (string-to-rational "   -123   ")))
  (is (= [-123 "   " 8] (string-to-rational "   -123.   ")))
  (is (= [(- -123 (/ 4 10)) "   " 9] (string-to-rational "   -123.4   ")))
  (is (= [(- -123 (/ 45 100)) "   " 10] (string-to-rational "   -123.45   ")))

  ;; trailing space only
  (is (= [1 "   " 1] (string-to-rational "1   ")))
  (is (= [1 "   " 2] (string-to-rational "1.   ")))
  (is (= [12 "   " 2] (string-to-rational "12   ")))
  (is (= [12 "   " 3] (string-to-rational "12.   ")))
  (is (= [123 "   " 3] (string-to-rational "123   ")))
  (is (= [123 "   " 4] (string-to-rational "123.   ")))
  (is (= [(+ 123 (/ 4 10)) "   " 5] (string-to-rational "123.4   ")))
  (is (= [(+ 123 (/ 45 100)) "   " 6] (string-to-rational "123.45   ")))

  (is (= [1 "+   " 1] (string-to-rational "1+   ")))
  (is (= [1 "+   " 2] (string-to-rational "1.+   ")))
  (is (= [1 " +   " 2] (string-to-rational "1. +   ")))
  (is (= [1 "   " 2] (string-to-rational "+1   ")))
  (is (= [1 "   " 3] (string-to-rational "+1.   ")))
  (is (= [12 "   " 3] (string-to-rational "+12   ")))
  (is (= [12 "   " 4] (string-to-rational "+12.   ")))
  (is (= [123 "   " 4] (string-to-rational "+123   ")))
  (is (= [123 "   " 5] (string-to-rational "+123.   ")))
  (is (= [(+ 123 (/ 4 10)) "   " 6] (string-to-rational "+123.4   ")))
  (is (= [(+ 123 (/ 45 100)) "   " 7] (string-to-rational "+123.45   ")))

  (is (= [1 "-   " 1] (string-to-rational "1-   ")))
  (is (= [1 "-   " 2] (string-to-rational "1.-   ")))
  (is (= [1 " -   " 2] (string-to-rational "1. -   ")))
  (is (= [-1 "   " 2] (string-to-rational "-1   ")))
  (is (= [-1 "   " 3] (string-to-rational "-1.   ")))
  (is (= [-12 "   " 3] (string-to-rational "-12   ")))
  (is (= [-12 "   " 4] (string-to-rational "-12.   ")))
  (is (= [-123 "   " 4] (string-to-rational "-123   ")))
  (is (= [-123 "   " 5] (string-to-rational "-123.   ")))
  (is (= [(- -123 (/ 4 10)) "   " 6] (string-to-rational "-123.4   ")))
  (is (= [(- -123 (/ 45 100)) "   " 7] (string-to-rational "-123.45   "))))

(defn -main [& args]
  (run-tests 'net.serice.clojure.rational-to-string-test))
